import logging

from rita.translate import rules_to_patterns
from rita.parser import RitaParser

logger = logging.getLogger(__name__)


def compile_string(raw):
    parser = RitaParser()
    parser.build()
    root = parser.parse(raw)
    logger.debug(root)
    result = [rule for doc in root if doc for rule in rules_to_patterns(doc())]
    return result


def compile(fname):
    with open(fname, "r") as f:
        raw = f.read()

    return compile_string(raw)



import logging

from itertools import chain
from importlib import import_module

logger = logging.getLogger(__name__)

VARIABLES = {}
MODULES = []


def flatten(lst):
    if len(lst) > 1:
        return lst

    def explode(v):
        if callable(v):
            return v()
        else:
            return v

    new_lst = map(explode, lst)
    return chain(*new_lst)


def resolve_value(obj, context):
    logger.debug("Resolving value: {0}, context: {1}".format(obj, context))
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, list):
        for item in obj:
            context.append(item)
        return context
    return obj(context=context)


def ANY(context, op=None):
    context.append(("regex", r".*", op))
    return context


def PUNCT(context, op=None):
    context.append(("punct", None, op))
    return context


def MARK(type_, obj, op=None):
    return {"label": resolve_value(type_, []), "data": resolve_value(obj, [])}


def LOAD(*args, context=None):
    fpath = resolve_value(args[0], {})
    with open(fpath, "r") as f:
        return list([l.strip() for l in f.readlines()])


def ASSIGN(k, v, context=None, op=None):
    logger.debug("Assigning: {0} -> {1}".format(k, v))
    VARIABLES[k] = resolve_value(v, [])


def IN_LIST(*args, context, op=None):
    variants = []
    new_context = []
    for arg in flatten(args):
        variants.append(resolve_value(arg, new_context))
    context.append(("any_of", variants, None))
    return context


def PATTERN(*args, context=None, op=None):
    new_ctx = []
    for arg in args:
        resolve_value(arg, new_ctx)
    return new_ctx


def WORD(*args, context, op=None):
    if len(args) == 1:
        literal = resolve_value(args[0], [])
        context.append(("value", literal, op))
        return context
    elif len(args) == 0:
        return context.append(("regex", "\\w+", op))


def NUM(*args, context, op=None):
    if len(args) == 1:
        literal = resolve_value(args[0], {})
        context.append(("value", literal, op))
        return context
    elif len(args) == 0:
        return context.append(("regex", "\\d+", op))


def POS(name, context, op=None):
    return context.append(("pos", resolve_value(name, {}), op))


def LEMMA(name, context, op=None):
    return context.append(("lemma", resolve_value(name, {}), op))


def ENTITY(name, context, op=None):
    return context.append(("entity", resolve_value(name, {}), op))


def IMPORT(module, context=None, op=None):
    mod_name = resolve_value(module, {})
    MODULES.append(import_module(mod_name))


def EXEC(obj, context=None, op=None):
    return resolve_value(obj, [])



import logging

from functools import partial

logger = logging.getLogger(__name__)


def any_of_parse(lst, op=None):
    d = {"TEXT": {"REGEX": r"({})".format("|".join(lst))}}
    if op:
        d["OP"] = op
    return d


def regex_parse(r, op=None):
    d = {"TEXT": {"REGEX": r}}

    if op:
        d["OP"] = op
    return d


def fuzzy_parse(r, op=None):
    # TODO: build premutations
    d = {"LOWER": {"REGEX": "({0})[.,?;!]?".format("|".join(r))}}
    if op:
        d["OP"] = op
    return d


def generic_parse(tag, value, op=None):
    d = {}
    d[tag] = value
    if op:
        d["OP"] = op
    return d


PARSERS = {
    "any_of": any_of_parse,
    "value": partial(generic_parse, "ORTH"),
    "regex": regex_parse,
    "entity": partial(generic_parse, "ENT_TYPE"),
    "lemma": partial(generic_parse, "LEMMA"),
    "pos": partial(generic_parse, "POS"),
    "punct": partial(generic_parse, "PUNCT"),
    "fuzzy": fuzzy_parse,
}


def rules_to_patterns(rule):
    if rule:
        logger.info(rule)
        yield {
            "label": rule["label"],
            "pattern": [PARSERS[t](d, op) for (t, d, op) in rule["data"]],
        }



import logging

import ply.yacc as yacc

from importlib import import_module
from functools import partial

from rita.lexer import RitaLexer
from rita import macros

logger = logging.getLogger(__name__)


def stub(*args, **kwargs):
    return None


def load_macro(name):
    try:
        return getattr(macros, name)
    except:
        pass

    def lazy_load(*args, **kwargs):
        for mod in macros.MODULES:
            try:
                fn = getattr(mod, name)
                return fn(*args, **kwargs)
            except Exception as ex:
                logger.error(ex)
                continue

        raise RuntimeError("MACRO {} not found".format(name))

    return lazy_load


def var_wrapper(variable):
    def wrapper(*args, **kwargs):
        logger.debug("Variables: {}".format(macros.VARIABLES))
        return macros.VARIABLES[variable]

    return wrapper


class RitaParser(object):
    tokens = RitaLexer.tokens
    precedence = (
        ("nonassoc", "ARROW"),
        ("nonassoc", "COMMA"),
        ("left", "ASSIGN"),
        ("left", "RBRACKET", "LBRACKET", "LPAREN", "RPAREN"),
        ("left", "KEYWORD", "NAME", "LITERAL"),
        ("right", "MODIF_QMARK", "MODIF_STAR", "MODIF_PLUS"),
    )

    def p_document(self, p):
        """
        DOCUMENT : MACRO_CHAIN
                 | VARIABLE
        """
        logger.debug("Building initial document {}".format(p[1]))
        p[0] = [p[1]]

    def p_document_list(self, p):
        """
        DOCUMENT : DOCUMENT MACRO_CHAIN
                 | DOCUMENT VARIABLE
        """
        logger.debug("Extending document {}".format(p[2]))
        p[0] = p[1] + [p[2]]

    def p_macro_chain(self, p):
        " MACRO_CHAIN : MACRO ARROW MACRO "
        logger.debug("Have {0} -> {1}".format(p[1], p[3]))
        p[0] = partial(p[3], p[1])

    def p_macro_chain_from_array(self, p):
        " MACRO_CHAIN : ARRAY ARROW MACRO "
        logger.debug("Have {0} -> {1}".format(p[1], p[3]))
        p[0] = partial(p[3], macros.PATTERN(*p[1]))

    def p_macro_chain_exec(self, p):
        " MACRO_CHAIN : EXEC MACRO "
        p[0] = partial(macros.EXEC, p[2])

    def p_macro_w_modif(self, p):
        """
        MACRO : MACRO MODIF_PLUS
              | MACRO MODIF_STAR
              | MACRO MODIF_QMARK
        """
        logger.debug("Adding modifier to Macro {}".format(p[1]))
        fn = p[1]
        p[0] = partial(fn, op=p[2])

    def p_macro_wo_args(self, p):
        " MACRO : KEYWORD "
        fn = load_macro(p[1])
        logger.debug("Parsing macro (w/o args): {}".format(p[1]))
        p[0] = fn

    def p_macro_w_args(self, p):
        " MACRO : KEYWORD LPAREN ARGS RPAREN "
        logger.debug("Parsing macro: {0}, args: {1}".format(p[1], p[3]))
        fn = load_macro(p[1])
        p[0] = partial(fn, *p[3])

    def p_macro_from_array(self, p):
        " MACRO : KEYWORD ARRAY "
        logger.debug("Parsing macro: {0}, args: {1}".format(p[1], p[2]))
        fn = load_macro(p[1])
        print("Args: {0} and {1}".format(p[1], p[2]))
        p[0] = partial(fn, *p[2])

    def p_array(self, p):
        " ARRAY : LBRACKET ARGS RBRACKET "
        p[0] = p[2]

    def p_variable(self, p):
        " VARIABLE_NAME : NAME "
        p[0] = var_wrapper(p[1])

    def p_variable_from_args(self, p):
        " VARIABLE : NAME ASSIGN ARGS "
        if len(p[3]) == 1:
            macros.ASSIGN(p[1], p[3][0])
        else:
            macros.ASSIGN(p[1], p[3])

    def p_arg_list(self, p):
        " ARGS : ARGS COMMA ARG "
        p[0] = p[1] + [p[3]]

    def p_args(self, p):
        " ARGS : ARG "
        p[0] = [p[1]]

    def p_arg(self, p):
        " ARG : LITERAL "
        p[0] = p[1]

    def p_arg_from_macro(self, p):
        " ARG : MACRO "
        p[0] = p[1]

    def p_arg_from_var(self, p):
        " ARG : VARIABLE_NAME "
        p[0] = p[1](context={})

    def p_arg_from_array(self, p):
        " ARGS : ARRAY "
        p[0] = p[1]

    def p_error(self, p):
        logger.error("Syntax error at '{}'".format(p.value))

    def build(self, **kwargs):
        self.lexer = RitaLexer().build(**kwargs)
        self.parser = yacc.yacc(module=self, errorlog=logger, **kwargs)

    def parse(self, data):
        return self.parser.parse(data, lexer=self.lexer, debug=logger)
